<?php

/**
 * Represents a Fusion server resource.
 */
class SearchApiFusionConnection extends SearchApiSolrConnection {

  /**
   * Path to the signals servlet.
   */
  const SIGNALS_SERVLET = 'signals';

  /**
   * Route search requests to the Query API (api/apps/<app>/query/<query-profile>).
   * Route signal requests to the Signals API (api/signals/<app>).
   * Route other requests to the Solr API (api/<app>/<handler>).
   */
  protected function constructUrl($servlet, array $params = array(), $added_query_string = NULL) {
    switch($servlet) {
      case self::SEARCH_SERVLET:
        // See https://doc.lucidworks.com/fusion-server/4.1/reference-guides/api/apps-api.html
        // See https://doc.lucidworks.com/fusion-server/4.1/reference-guides/api/query-api.html
        $servlet = 'apps/' . $this->options['fusion_app'] . '/query/' . $this->options['fusion_qprofile_search'];
        break;
      case self::SIGNALS_SERVLET:
        // See https://doc.lucidworks.com/fusion-ai/4.1/reference-guides/api/signals-api.html
        $servlet = 'signals/' . $this->options['fusion_app'];
        break;
      default:
        // See https://doc.lucidworks.com/fusion-server/4.1/reference-guides/api/solr-api.html
        $servlet = 'solr/' . $this->options['fusion_app'] . '/' . $servlet;
    }
    return parent::constructUrl($servlet, $params, $added_query_string);
  }

  /**
   * Get autocomplete suggestions.
   *
   * @param $searchQuery
   *   The search query.
   *
   * @see SearchApiAutocompleteSuggesterInterace::getAutocompleteSuggestions().
   */
  public function getAutocompleteSuggestions($searchQuery) {
    // Send request to Fusion.
    // search_api_page shows a maximum of 10 suggestions, so we request only 10 entries from Fusion.
    $servlet = 'apps/' . $this->options['fusion_app'] . '/query/' . $this->options['fusion_qprofile_autocomplete'];
    $params = array(
      'start' => 0,
      'rows' => 10,
      'q' => $searchQuery,
      'sort' => '',
    );

    $url = parent::constructurl($servlet, $params);
    $response = $this->sendRawGet($url);

    // Log response if necessary.
    if (!empty($this->options['log_query'])) {
      $this->logResponse($response);
    }

    // Parse response and return an array of suggestions.
    $suggestions = array();
    foreach ($response->response->docs as $doc) {
      $suggestions[] = $doc->query; // No need to sanitize (search_api_autocomplete takes care of this).
    }

    return $suggestions;
  }

  /**
   * Send a signal to Fusion.
   *
   * @param $type
   *   Type of signal.
   * @param $params
   *   Signal parameters.
   * @param $timeout
   *   Timeout when sending request to Fusion.
   *   We do not want signal requests to slow down too much the user experience, in case Fusion happens to be slow.
   *   As a result we set a low timeout (5 seconds instead of 30 seconds by default).
   *
   * @see https://doc.lucidworks.com/fusion-ai/4.0/user-guide/signals/signals.html#built-in-signal-types
   */
  public function sendSignal($type, $params, $timeout = 5) {
    // Build URL and body of the request.
    // We could populate the session parameter with the value returned by session_id(), although this would be useful
    // only for authenticated users as this function returns a different value on each page load for anonymous users.
    $url = $this->constructUrl(self::SIGNALS_SERVLET);
    $data = array(
      array(
        'timestamp' => intval(microtime(TRUE) * 1000),
        'type' => $type,
        'params' => $params + array(
          'app_id' => $this->options['server'],
          'ip_address' => ip_address(),
          'referrer' => $_SERVER['HTTP_REFERER'],
        ),
      )
    );
    $options = array(
      'method' => 'POST',
      'headers' => array(
        'Content-Type' => 'application/json',
      ),
      'timeout' => $timeout,
      'data' => json_encode($data),
    );

    // Send request.
    // We do not use $this->sendRawPost(), as the latter throws an exception when the response is not HTTP 200 (Fusion returns 204).
    $response = $this->makeHttpRequest($url, $options);

    // Verify response and return.
    if ($response->code != 204) {
      throw new SearchApiException("Signal sent to $url returned HTTP code $response->code. Data posted: " . json_encode($data));
    }
    return $response;
  }

}
