<?php

/**
 * Sarnia Search service class for Fusion.
 *
 * Use it when the Solr schema installed in Fusion is *not* provided by search_api_solr.
 */
class SarniaFusionService extends SarniaSolrService {

  /**
   * Use Fusion connection class.
   */
  protected $connection_class = 'SearchApiFusionConnection';

  /**
   * Use Fusion trait.
   */
  use FusionTrait;

  /**
   * {@inheritdoc}
   */
  protected function preQuery(array &$call_args, SearchApiQueryInterface $query) {
    // Set the default query and request handler via sarnia.
    parent::preQuery($call_args, $query);

    // Remove qf parameter set by SearchApiSolrConnection as we don't need this when querying Fusion.
    unset($call_args['params']['qf']);

    // Highlighting is also configured on Fusion itself so there is no need to send hl parameters.
    foreach ($call_args['params'] as $param => $value) {
      if (substr($param, 0, strlen('hl')) === 'hl') {
        unset($call_args['params'][$param]);
      }
    }

    // Spellchecker is also configured on Fusion itself.
    unset($call_args['params']['spellcheck']);

    // Sarnia sets the fl parameter to '*' which causes all fields to be loaded on Fusion and is bad for performance.
    // Hence we remove it.
    // As a result if you build a view based on Fusion you have to use the fields returned by default by Fusion.
    unset($call_args['params']['fl']);
  }

}

/**
 * Solr Search service class for Fusion.
 *
 * Use it when the Solr schema installed in Fusion *is* provided by search_api_solr.
 */
class SolrFusionService extends SearchApiSolrService {

  /**
   * Use Fusion connection class.
   */
  protected $connection_class = 'SearchApiFusionConnection';

  /**
   * Use Fusion trait.
   */
  use FusionTrait;

}

/**
 * Fusion trait common to both service classes (SarniaFusionService, SolrFusionService).
 */
trait FusionTrait {

  /**
   * {@inheritdoc}
   */
  public function configurationForm(array $form, array &$form_state) {
    // Get original form.
    $form = parent::configurationForm($form, $form_state);

    // Change field labels and default values (from Solr to Fusion).
    if ($this->options) {
      $form['server_description']['#title'] = t('Fusion server URI');
    }

    $options = $this->options + array(
      'port' => '8764',
      'retrieve_data' => TRUE,
      'skip_schema_check' => TRUE,
      'fusion_app' => '',
      'fusion_qprofile_search' => '',
      'fusion_qprofile_autocomplete' => '',
      'fusion_excerpt_field' => '',
      'fusion_signals' => FALSE,
    );

    $form['host']['#title'] = t('Fusion host');
    $form['host']['#description'] = t('The host name or IP of your Fusion server, e.g. <code>localhost</code> or <code>www.example.com</code>.');

    $form['port']['#title'] = t('Fusion port');
    $form['port']['#description'] = t('By default the Fusion UI runs on port 8764.');
    $form['port']['#default_value'] = $options['port'];

    $form['advanced']['retrieve_data']['#default_value'] = $options['retrieve_data'];
    $form['advanced']['skip_schema_check']['#default_value'] = $options['skip_schema_check'];

    // Make the path static and set it to /api.
    $form['path'] = array(
      '#type' => 'value',
      '#value' => '/api',
    );

    // Add extra fields specific to Fusion.
    $form['fusion_app'] = array(
      '#type' => 'textfield',
      '#title' => t('Fusion app'),
      '#description' => t('The name of the Fusion app.'),
      '#default_value' => $options['fusion_app'],
      '#required' => TRUE,
    );
    $form['fusion_qprofile_search'] = array(
      '#type' => 'textfield',
      '#title' => t('Fusion search query profile'),
      '#description' => t('The Fusion query profile to use when searching.'),
      '#default_value' => $options['fusion_qprofile_search'],
      '#required' => TRUE,
    );
    if (module_exists('search_api_autocomplete')) {
      $form['fusion_qprofile_autocomplete'] = array(
        '#type' => 'textfield',
        '#title' => t('Fusion autocomplete query profile'),
        '#description' => t('The Fusion query profile to use when generating autocomplete suggestions.'),
        '#default_value' => $options['fusion_qprofile_autocomplete'],
        '#required' => TRUE,
      );
    }
    $form['fusion_excerpt_field'] = array(
      '#type' => 'textfield',
      '#title' => t('Field used for excerpts'),
      '#description' => t('The field containing the search result excerpts.'),
      '#default_value' => $options['fusion_excerpt_field'],
    );
    $form['fusion_signals'] = array(
      '#type' => 'checkbox',
      '#title' => 'Send signals to Fusion',
      '#description' => t('<a href="@url_request">Request signals</a> are sent after search results are returned. <a href="@url_click">Click signals</a> are sent after users click on a search result.', array('@url_request' => 'https://doc.lucidworks.com/fusion-ai/4.1/user-guide/signals/signals.html#request', '@url_click' => 'https://doc.lucidworks.com/fusion-ai/4.1/user-guide/signals/signals.html#click')),
      '#default_value' => $options['fusion_signals'],
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtraInformation() {
    // Show original information.
    $info = parent::getExtraInformation();

    // Show extra fields specific to Fusion.
    $info[] = array(
      'label' => t('Fusion app'),
      'info' => $this->options['fusion_app'],
    );
    $info[] = array(
      'label' => t('Fusion search query profile'),
      'info' => $this->options['fusion_qprofile_search'],
    );
    $info[] = array(
      'label' => t('Fusion autocomplete query profile'),
      'info' => $this->options['fusion_qprofile_autocomplete'],
    );
    $info[] = array(
      'label' => t('Send signals to Fusion'),
      'info' => $this->options['fusion_signals'] ? t('Yes') : t('No'),
    );

    return $info;
  }


  /**
   * {@inheritdoc}
   */
  protected function postQuery(array &$results, SearchApiQueryInterface $query, $response) {
    parent::postQuery($results, $query, $response);

    // Send Request signal after search response was received (if signals are enabled).
    if (!empty($this->options['fusion_signals']) && user_access('send signals to any fusion server')) {
      $this->sendRequestSignal($query, $results);
    }
  }

  /**
   * See Request signal to Fusion.
   *
   * @param $query
   *   Search API query object.
   * @param $results
   *   Array of search results received from Fusion.
   *
   * @see https://doc.lucidworks.com/fusion-ai/4.1/user-guide/signals/signals.html#request
   */
  public function sendRequestSignal(SearchApiQueryInterface $query, $results) {
    $keys = $query->getOriginalKeys();
    if (is_array($keys)) {
      $keys = $query->flattenKeys($keys);
    }
    $params = $this->getSignalParamsFilter($query) + array(
      'query' => $keys,
      'page_title' => drupal_get_title(),
    );
    try {
      $this->connect();
      return $this->solr->sendSignal('request', $params);
    }
    catch (SearchApiException $e) {
      watchdog_exception('search_api_fusion', $e, '%type while sending request signal to Fusion: !message in %function (line %line of %file)', array(), WATCHDOG_WARNING);
    }
    return array();
  }

  /**
   * Send Click signal to Fusion.
   *
   * @param $params
   *   Array of parameters as per https://doc.lucidworks.com/fusion-ai/4.1/user-guide/signals/signals.html#click
   */
  public function sendClickSignal($params) {
    try {
      $this->connect();
      $response = $this->solr->sendSignal('click', $params);
    }
    catch (SearchApiException $e) {
      watchdog_exception('search_api_fusion', $e, '%type while sending click signal to Fusion: !message in %function (line %line of %file)', array(), WATCHDOG_WARNING);
    }
    return $response;
  }

  /**
   * {@inheritdoc}
   */
  protected function extractResults(SearchApiQueryInterface $query, $response) {
    $results = parent::extractResults($query, $response);

    // Get original search query.
    $keys = $query->getOriginalKeys();
    if (is_array($keys)) {
      $keys = $query->flattenKeys($keys);
    }

    // Get server name.
    $server_name = $query->getIndex()->server;

    // Get Fusion Query ID.
    $fusion_query_id = $results['search_api_solr_response']->headers['x-fusion-query-id'];

    if (!empty($this->options['fusion_signals'])) {
      // Get signal parameters for filters.
      $params_filter = $this->getSignalParamsFilter($query);
    }

    // Extract landing pages.
    if (!empty($response->fusion->{'landing-pages'})) {
      $landing_pages = $response->fusion->{'landing-pages'};
      foreach ($landing_pages as $landing_page) {
        list($url, $title) = explode('$$$', $landing_page);
        $results['fusion_landing_pages'][$url] = array(
          'title' => $title,
          'url' => array(
            'path' => $url,
            'options' => array('absolute' => TRUE, 'external' => TRUE),
          ),
        );
        if (!empty($this->options['fusion_signals'])) {
          // Landing pages do not seem to require res_pos to be set.
          $params = $params_filter + array(
            'query' => $keys,
            'ctype' => 'landingpage',
            'fusion_query_id' => $fusion_query_id,
            'doc_id' => $url,
            'doc_ids_s' => $url,
            'url' => $url,
          );
          $results['fusion_landing_pages'][$url]['url']['options']['attributes']['ping'] = $this->constructClickSignalUrl($server_name, $params);
        }
      }
    }

    // Fusion outputs spellcheck suggestions in a slightly different way than Solr so we use a custom class.
    if (module_exists('search_api_spellcheck') && $query->getOption('search_api_spellcheck')) {
      $results['search_api_spellcheck'] = new SearchApiSpellcheckFusion($response);
    }

    // Populate results with URL of click signal (if signals are enabled).
    if (!empty($this->options['fusion_signals'])) {
      $start = $results['search_api_solr_response']->response->start;
      $local_pos = 0;
      foreach ($results['results'] as $id => $result) {
        $local_pos++;
        $params = $params_filter + array(
          'query' => $keys,
          'ctype' => 'result',
          'res_pos' => $start + $local_pos,
          'fusion_query_id' => $fusion_query_id,
          'doc_id' => $id,
          'doc_ids_s' => $id,
          'url' => $id,
        );
        $results['results'][$id]['ping'] = $this->constructClickSignalUrl($server_name, $params);
      }
    }

    return $results;
  }

  /**
   * Build URL of the proxy to send a click signal.
   *
   * @param $server_name
   *   Machine name of the Search API Server.
   * @param $params
   *   Array of parameters as per https://doc.lucidworks.com/fusion-ai/4.1/user-guide/signals/signals.html#click
   *
   * @see search_api_fusion_send_click_signal().
   */
  protected function constructClickSignalUrl($server_name, $params) {
    return url("search_api_fusion/signals/click/$server_name", array('query' => $params));
  }

  /**
   * Generate signal parameters "filter" and "filter_field" based on which facets are active.
   *
   * @param $query
   *   Search query.
   *
   * @return
   *   Array with two keys: "filter" and "filter_field" with values expected by Fusion.
   *   Empty array if no facet is active.
   *
   * @see https://doc.lucidworks.com/fusion-ai/4.0/user-guide/signals/signals.html#important-fields-for-signals
   */
  protected function getSignalParamsFilter(SearchApiQueryInterface $query) {
    $params = array();
    $filters = $query->getFilter()->getFilters();
    foreach ($filters as $filter) {
      foreach ($filter->getFilters() as $condition) {
        if (!is_array($condition)) {
          // Fusion's signals API has no documentation about how to deal with complex filters (e.g. [X AND Y] OR Z) so we just ignore those.
          continue;
        }
        list($field, $value, $operator) = $condition;
        if ($operator != '=') {
          // Fusion's signals API has no documentation about how to deal with operators other than '=' so we just ignore those.
          continue;
        }
        $params['filter_field'][] = $field;
        if ($value === NULL) {
          // Support facets for missing values.
          // Those generate an fq made of two conditions and a minus sign e.g. *:* AND -myfield:[* TO *]
          // @see SearchApiSolrService::createFilterQuery()
          // Fusion's signals API does not support filters made of multiple conditions and requires those to be formatted like this: myfield/myvalue
          // @see https://doc.lucidworks.com/fusion-ai/4.0/user-guide/signals/signals.html#important-fields-for-signals
          // As a workaround we use the bang value (!) which is used by Drupal and seems to work with Fusion.
          // @see SearchApiSolrService::extractFacets()
          // Response events do show up with the two conditions, albeit with a missing parenthesis e.g. (*:* AND -myfield:[* TO *]
          $params['filter'][] = "$field/!";
        }
        else {
          // Sarnia wraps $value with double quotes: we strip those out for better readability.
          // It is not necessary to escape slashes in $value -- Fusion will properly parse the value.
          $params['filter'][] = $field . '/' . trim($value, '"');
        }
      }
    }
    return $params;
  }

  /**
   * {@inheritdoc}
   */
  public function getAutocompleteSuggestions(SearchApiQueryInterface $query, SearchApiAutocompleteSearch $search, $incomplete_key, $user_input) {
    $suggestions = array();

    try {
      $this->connect();
      $suggestions = $this->solr->getAutocompleteSuggestions($user_input);
    }
    catch (SearchApiException $e) {
      watchdog_exception('search_api_fusion', $e, '%type during autocomplete Fusion query: !message in %function (line %line of %file).', array(), WATCHDOG_WARNING);
    }

    return $suggestions;
  }

  /**
   * {@inheritdoc}
   */
  protected function getExcerpt($response, $id, array &$fields, array $field_mapping, array $fulltext_fields) {
    // Extract excerpts, if a field was provided. This is used by search_api_page.
    // TODO turn into a drop-down list instead of a text field.
    if (!empty($this->options['fusion_excerpt_field'])) {
      $excerpt_field = $this->options['fusion_excerpt_field'];
      if (isset($fields[$excerpt_field])) {
        return $fields[$excerpt_field];
      }
    }
  }

}
